#!/usr/sepp/bin/perl -w

# mailgraph -- a postfix statistics rrdtool frontend
# copyright (c) 2000, David Schweikert <dws@ee.ethz.ch>
# released under the GNU General Public License

use lib "/usr/pack/rrdtool-1.0.17-za/lib/perl";
use lib "/home/dws/lib/site_perl/5.6.0";
use lib "/home/dws/lib/site_perl/5.6.0/sun4-solaris/";

use RRDs;
use Time::Local;

use strict;
use File::Tail;

my $VERSION = 0.6;

# config
my $rrdstep = 60;
my $xpoints = 540;
my $points_per_sample = 3;
my $ypoints = 150;
my $ypoints_err = 80;
my $graph_period = 1800; # 30 minutes
my $graph_mintime = 20; # 20 seconds

# global variables
my $logfile;
my $dir; # where to store rrd and graphs
my $rrd; # rrd file
my $year;
my $this_minute;
my %sum = ( sent => 0, received => 0, bounced => 0, rejected => 0 );
my $rrd_inited=0;
my $syslog_re = qr'^(...)[ ]{1,2}(\d+) (\d+):(\d+):(\d+) (\S+) (\S+?): (.*)'o;
my $last_graphed=0;

# prototypes
sub process_line($);
sub syslogtime2time($$$$$);
sub event_sent($);
sub event_received($);
sub event_bounced($);
sub event_rejected($);
sub init_rrd($);
sub update($);
sub graph($);

sub main
{
	print "mailgraph $VERSION by dws\@ee.ethz.ch\n";

	my $catfile=0;
	$logfile = shift @ARGV;
	if(defined $logfile and $logfile eq '-c') {
		$catfile=1;
		$logfile = shift @ARGV;
	}
	$dir = shift @ARGV;
	defined $dir or do {
		print "usage: mailgraph [-c] logfile dir\n";
		print "-c causes the logfile to be only read and not monitored\n";
		exit;
	};
	$rrd = "$dir/mailgraph.rrd";

	if(not $catfile) {
		my $file = File::Tail->new(name=>$logfile, tail=>-1);
		my $line;
		my $graph_minute=0;
		my $graph_time=0;
		while (defined($line=$file->read)) {
			process_line($line);
			$this_minute or next;
			if($this_minute-$graph_minute >= $graph_period) {
				my $now = time;
				next if $now-$graph_time<$graph_mintime;
				graph($this_minute);
				$graph_minute=$this_minute;
				$graph_time = $now;
			}
		}
	}
	else {
		open(FILE, "<$logfile") or die "can't open $logfile\n";
		while(<FILE>) {
			process_line($_);
		}
		graph($this_minute);
	}
}

sub init_rrd($)
{
	my $m = shift;
	if(not -f $rrd) {
		my $rows = $xpoints/$points_per_sample;
		my $realrows = int($rows*1.5); # ensure that the full range is covered
		my $day_steps = int(3600*24 / ($rrdstep*$rows));
		my $week_steps = $day_steps*7;
		my $month_steps = $day_steps*31;
		my $year_steps = $day_steps*365;
		RRDs::create($rrd, '--start', $m, '--step', $rrdstep,
				'DS:sent:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:recv:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:bounced:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:rejected:ABSOLUTE:'.($rrdstep*2).':0:U',
				"RRA:AVERAGE:0.5:$day_steps:$realrows",   # day
				"RRA:AVERAGE:0.5:$week_steps:$realrows",  # week
				"RRA:AVERAGE:0.5:$month_steps:$realrows", # month
				"RRA:AVERAGE:0.5:$year_steps:$realrows",  # year
				);
		$this_minute = $m;
	}
	else {
		$this_minute = RRDs::last($rrd) + $rrdstep;
	}
	$rrd_inited=1;
}

sub syslogtime2time($$$$$)
{

	my ($mstr,$d,$H,$M,$S) = @_;
	my %months = (
		'Jan' => 0, 'Feb' =>  1, 'Mar' =>  2,
		'Apr' => 3, 'May' =>  4, 'Jun' =>  5,
		'Jul' => 6, 'Aug' =>  7, 'Sep' =>  8,
		'Oct' => 9, 'Nov' => 10, 'Dec' => 11 );
	my $m = $months{$mstr} or do {
		warn "Can't parse month '$1'\n";
		return;
	};

	$year = (localtime(time))[5] unless defined $year;
	my $time = timelocal($S,$M,$H,$d,$m,$year);
	if(defined $this_minute and $time < $this_minute) {
		# Dec 31 -> Jan  1 
		$year = (localtime(time))[5];
		$time = timelocal($S,$M,$H,$d,$m,$year);
	}
	return $time;
}

sub process_line($)
{
	my $line = shift;
	# parse syslog syntax
	$line =~ /$syslog_re/o or return;
	my ($m,$d,$H,$M,$S) = ($1,$2,$3,$4,$5);
	my $host = $6;
	my $prog = $7;
	my $str = $8;
	$prog =~ /^postfix\/(.*)/ or return; $prog = $1;
	my $time = syslogtime2time($m,$d,$H,$M,$S);
	$prog =~ s/\[\d+\]//; # strip process number
	$prog eq 'smtp' and do {
		if($str =~ /\bstatus=sent\b/) {
			event_sent($time);
		}
		return; };
	$prog eq 'local' and do {
		if($str =~ /\bstatus=bounced\b/) {
			event_bounced($time);
		}
		return; };
	$prog eq 'smtpd' and do {
		if($str =~ /^[0-9A-F]{10}: client=/) {
			event_received($time);
		}
		if($str =~ /^reject: /) {
			event_rejected($time);
		}
		return; };
}

sub event_sent($)
{
	my $t = shift;
	update($t) and $sum{sent}++;
}

sub event_received($)
{
	my $t = shift;
	update($t) and $sum{received}++;
}

sub event_bounced($)
{
	my $t = shift;
	update($t) and $sum{bounced}++;
}

sub event_rejected($)
{
	my $t = shift;
	update($t) and $sum{rejected}++;
}

# returns 1 if $sum should be updated
sub update($)
{
	my $t = shift;
	my $m = $t - $t%$rrdstep;
	init_rrd($m) unless $rrd_inited;
	return 1 if $m == $this_minute;
	return 0 if $m < $this_minute;

	#print "$m\n";
	print "update $this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}\n";
	RRDs::update $rrd, "$this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}";
	if($m > $this_minute+$rrdstep) {
		for(my $sm=$this_minute+$rrdstep;$sm<$m;$sm+=$rrdstep) {
			print "update $sm:0:0:0:0 (SKIP)\n";
			RRDs::update $rrd, "$sm:0:0:0:0";
		}
	}
	$this_minute = $m;
	$sum{sent}=0;
	$sum{received}=0;
	$sum{bounced}=0;
	$sum{rejected}=0;
	return 1;
}

sub make_graph($$$$)
{
	my $now = shift;
	my $range = shift;
	my $file = shift;
	my $title = shift;

	RRDs::graph("$file.png", '-a', 'PNG', '-w', $xpoints, '-h', $ypoints,
		'-s', $now-$range, '-e', $now,
		'-v', 'msgs/min', '-t', $title,
		"DEF:sent=$rrd:sent:AVERAGE",
		"DEF:recv=$rrd:recv:AVERAGE",
		"CDEF:rsent=sent,60,*",
		"CDEF:rrecv=recv,60,*",
		"CDEF:vsent=sent,$range,*",
		"CDEF:vrecv=recv,$range,*",
		'AREA:rsent#000099:Sent',
		'LINE2:rrecv#009900:Received',
		'GPRINT:vsent:AVERAGE:Sent\: %.0lf msgs',
		'GPRINT:vrecv:AVERAGE:Received\: %.0lf msgs',
	);
	not RRDs::error or die "ERROR while graphing $file: ".RRDs::error."\n";

	RRDs::graph("${file}_err.png", '-a', 'PNG', '-w', $xpoints, '-h', $ypoints_err,
		'-s', $now-$range, '-e', $now,
		'-v', 'msgs/hour', '-t', "$title of Errors",
		"DEF:bounced=$rrd:bounced:AVERAGE",
		"DEF:rejected=$rrd:rejected:AVERAGE",
		"CDEF:rbounced=bounced,3600,*",
		"CDEF:rrejected=rejected,3600,*",
		"CDEF:vbounced=bounced,$range,*",
		"CDEF:vrejected=rejected,$range,*",
		'AREA:rbounced#000000:Bounced',
		'LINE2:rrejected#BB0000:Rejected',
		'GPRINT:vbounced:AVERAGE:Bounced\: %.0lf msgs',
		'GPRINT:vrejected:AVERAGE:Rejected\: %.0lf msgs',
	);
	not RRDs::error or die "ERROR while graphing $file: ".RRDs::error."\n";

	open(LAST,">$dir/mailgraph_lastupdated.txt");
	print LAST localtime($this_minute-$rrdstep)."\n";
	close(LAST);
}

sub graph($)
{
	my $now = shift;

	print "generating graphs...\n";

	my $day = 3600*24;
	my $week = $day*7;
	my $month = $day*31;
	my $year = $day*365;

	make_graph($now, $day, "$dir/mailgraph_day", "Day Graph");
	make_graph($now, $week, "$dir/mailgraph_week", "Week Graph");
	make_graph($now, $month, "$dir/mailgraph_month", "Month Graph");
	make_graph($now, $year-$day, "$dir/mailgraph_year", "Year Graph");
}

main;
